
#include <sys/types.h>
#include <sys/mman.h>
#include <stdio.h>
#include <assert.h>

#include "jiti-dr.h"

#define __MR__calloc__ calloc
#define __MR__free__   free

/* NIET WIJZIGEN!!! Elk blok moet 2k groot zijn. */
#define a1_bits   9
#define a2_bits   9
#define a3_bits   32-a1_bits-a2_bits

#define a1_aantal (1<<a1_bits)
#define a2_aantal (1<<a2_bits)
#define a3_aantal (1<<a3_bits)

bm_par  _MR_bms[MAXTHREADS];

/*
               +------+------+
 _MR_BMS---->  |      |      |
               +------+------+
               |      |      |
               +------+------+
               |      |      |  <-- thread_id
               .      .      .
               +------+------+
                  ^       ^
                  |       |
                load    store 
*/

unsigned  _MR_bmsize=sizeof(_MR_bms);  /* initiele grootte van de bitmaps */
unsigned  _MR_bmtotal=sizeof(_MR_bms);  /* initiele grootte van de bitmaps */

bitmap  _MR_combine_bitmaps(bitmap, bitmap);

extern unsigned _MR_pagesize;
extern int _MR_devzero;

/**************************************************/

void _MR_print_bitmap(bitmap bm){
  unsigned vp1,vp2,vp3,vp4;
  unsigned last=0;
  unsigned h,i,j,k,l,m;
  
  vp1=(unsigned)bm;
  if (vp1){
    printf("Bitmap at 0x%x\n", (unsigned)vp1);
    for (j=0; j<a1_aantal; j++){
      vp2 = (unsigned)vp1+j*4;	
      if (*(unsigned*)vp2){
	DBG(printf(" Superpage %x at 0x%x\n", j, *(unsigned*)vp2));
	for (k=0; k<a2_aantal; k++){
	  vp3=*(unsigned*)vp2+k*4;
	  if (*(unsigned*)vp3){
	    DBG(printf("  Page %x at 0x%x\n", k, *(unsigned*)vp3));		
	    for (l=0; l<a3_aantal/8; l+=4){
	      vp4=*(unsigned*)vp3+l;
	      if (*(unsigned*)vp4){
		for (m=0;m<32;m++)
		  if (*(unsigned*)vp4 & (1<<m))
		    printf("%8x ",(j<<22)+(k<<12)+l*8+m);
	      }
	    }
	    printf("\n");		
	  }
	}
      }
    }
  }
}


/**************************************************/

void  _MR_add_to_bitmap(unsigned address, unsigned bytes,
			unsigned thread_id, unsigned type){
  unsigned a1,a2,a3, a3_word,a3_bit;
  char * vp1, *vp2;
  unsigned * bitmap_32;

  a1 = address >> (a2_bits + a3_bits);
  a2 = (address >> a3_bits) & ((1<<a2_bits)-1);
  a3 = address & ((1<<a3_bits)-1);
  
  DBG(printf("t=%d type=%d a=0x%x a1=0x%x a2=0x%x a3=0x%x\n", 
	     thread_id, type, address, a1,a2,a3));
	     
  if (!  (vp1 = _MR_bms[thread_id].bm[type])  ){
    _MR_bms[thread_id].bm[type] = (bitmap)__MR__calloc__(a1_aantal, sizeof(void*));
    vp1 = _MR_bms[thread_id].bm[type];
    DBG(printf("Malloc for thread %d at 0x%x\n", thread_id, vp1));
  }
  
  vp1 = vp1+a1*4;
  if (!  (*(unsigned*)vp1)  ){ 
    *(void **)vp1 = (char*)__MR__calloc__(a2_aantal, sizeof(void *));
    DBG(printf("Malloc for thread %d, superpage %x at 0x%x\n", 
	       thread_id, a1, *(unsigned*)vp1));
  }
  
  vp1 = (char*)(*(unsigned*)vp1+a2*4);
  if (!  (*(unsigned*)vp1)  ){ 
    *(void **)vp1 = (char*)__MR__calloc__(a3_aantal/8, sizeof(char));
    DBG(printf("Malloc for thread %d, superpage %x, page %x at 0x%x\n", 
	       thread_id, a1, a2, *(unsigned*)vp1));
  }  
  
  a3_word = a3 / 32;
  bitmap_32 = (unsigned*)(*(unsigned*)vp1+a3_word*4);
  if (*bitmap_32==0xFFFFFFFF) return;

  a3_bit  = ((1<<bytes)-1) << (a3 % 32);
 *bitmap_32 |= a3_bit;
  
}

/**************************************************/

int _MR_compare_bitmaps(bitmap bm1, bitmap bm2){
  unsigned vp1_1,vp1_2,vp1_3,vp1_4;
  unsigned vp2_1,vp2_2,vp2_3,vp2_4;
  unsigned last=0;
  unsigned h,i,j,k,l,m;
  unsigned temp;

  DBG(printf("COMPARING %x with %x\n", bm1, bm2));

  if ((unsigned)bm1 & (unsigned)bm2){
    DBG(printf(" Cmp 0x%x - 0x%x\n", bm1, bm2));
    vp1_1 = (unsigned)bm1;
    vp2_1 = (unsigned)bm2;
    for (j=0; j<a1_aantal; j++){
      vp1_2 = (unsigned)vp1_1+j*4;	
      vp2_2 = (unsigned)vp2_1+j*4;
      if (*(unsigned*)vp1_2 & *(unsigned*)vp2_2){
	DBG(printf("  Checking superpage %x\n", j));
	for (k=0; k<a2_aantal; k++){
	  vp1_3=*(unsigned*)vp1_2+k*4;
	  vp2_3=*(unsigned*)vp2_2+k*4;
	  if (*(unsigned*)vp1_3 & *(unsigned*)vp2_3){
	    DBG(printf("   Checking page %x\n", k));
	    for (l=0; l<a3_aantal/8; l+=4){
	      vp1_4=*(unsigned*)vp1_3+l;
	      vp2_4=*(unsigned*)vp2_3+l;
	      temp = *(unsigned*)vp1_4 & *(unsigned*)vp2_4;
	      if (temp){
		for (m=0;m<32;m++)
		  if (temp & (1<<m)) {
		    DBG(printf(" <<< %8x >>> ",(j<<(a2_bits+a3_bits))+(k<<a3_bits)+l*8+m));
		    return (j<<(a2_bits+a3_bits))+(k<<a3_bits)+l*8+m;
		  }
	      }
	    }
	    DBG(printf("\n"));
	  }
	}
      }
    }
  }
  return 0;
}

/*************************************/

bitmap _MR_combine_bitmaps(bitmap bm1, bitmap bm2){
  unsigned vp1_1,vp1_2,vp1_3,vp1_4; /* bm1 */
  unsigned vp2_1,vp2_2,vp2_3,vp2_4; /* bm2 */
  unsigned vp3_1,vp3_2,vp3_3,vp3_4; /* out */
  unsigned last=0;
  unsigned h,i,j,k,l,m;
  unsigned temp;
  
  vp3_1 = (unsigned)NULL;
  
#define NEXT(x,y) ( (x) ? *(unsigned*)((unsigned)(x)+(y)*4) : NULL )
  
  if ((unsigned)bm1 || (unsigned)bm2){
    DBG(printf(" Combine 0x%x - 0x%x\n", bm1, bm2));
    vp1_1 = (unsigned)bm1;
    vp2_1 = (unsigned)bm2;
    for (j=0; j<a1_aantal; j++){
      vp1_2 = (unsigned)NEXT(vp1_1, j);
      vp2_2 = (unsigned)NEXT(vp2_1, j);
      vp3_2 = (unsigned)vp3_1+j*4;
      if (vp1_2 || vp2_2){
	if (!vp3_1){
	  vp3_1 = __MR__calloc__(a1_aantal, sizeof(void*));
	  vp3_2 = (unsigned)vp3_1+j*4;
	}
	*(void **)vp3_2 = (char*)__MR__calloc__(a2_aantal, sizeof(void *));
	DBG(printf("  Combining superpage %x\n", j));
	for (k=0; k<a2_aantal; k++){
	  vp1_3 = (unsigned)NEXT(vp1_2, k);
	  vp2_3 = (unsigned)NEXT(vp2_2, k);
	  vp3_3 = *(unsigned*)vp3_2+k*4;
	  if (vp1_3 || vp2_3){
	    DBG(printf("   Combining page %x\n", k));
	    *(void **)vp3_3 = (char*)__MR__calloc__(a3_aantal, sizeof(void *));
	    for (l=0; l<a3_aantal/8; l+=4){
	      vp1_4 = (unsigned)NEXT(vp1_3, l);
	      vp2_4 = (unsigned)NEXT(vp2_3, l);
	      vp3_4=*(unsigned*)vp3_3+l;
	       *(unsigned*)vp3_4 = vp1_4 | vp2_4;
	    }
	  }
	}
      }
    }
  }
  return (bitmap)vp3_1;
}  


/*************************************/

void _MR_remove_bitmap(bitmap bm){
  unsigned vp1,vp2,vp3,vp4;
  unsigned last=0;
  unsigned h,i,j,k,l,m;

  vp1 = (unsigned)bm;
  if (vp1){
    DBG(printf(" Removing bitmap %x\n", bm));
    for (j=0; j<a1_aantal; j++){
      vp2 = (unsigned)vp1+j*4;	
      if (*(unsigned*)vp2){
	DBG(printf("  Removing superpage %x\n", j));
	for (k=0; k<a2_aantal; k++){
	  vp3=*(unsigned*)vp2+k*4;
	  if (*(unsigned*)vp3){
	    DBG(printf("   Removing page %x\n", k));
	    __MR__free__((void*)*(unsigned*)vp3);
	  }
	}
	__MR__free__((void*)*(unsigned*)vp2);
      }
    }
    __MR__free__((void*)vp1);
  }
}  

/*************************************/
