
//extern unsigned int _MR_bmsize;

#define STOP_AFTER_FIRST_DR

#define MAX_SEGMENTS 20000

/******************************/

#define NORMAL_UPDATE 0
#define SEMA_P_UPDATE 1
#define SEMA_V_UPDATE 2

#include <stdio.h>
#include <assert.h>
#include <dlfcn.h>

#include "jiti-dr.h"

#define PREFIX "prefix_ontbreekt_nog."

unsigned  _MR_nr_cpus=MAXTHREADS;

segment_info  _MR_segment_list[MAX_SEGMENTS];

segment_info  _MR_segment_now[MAXTHREADS];

pthread_mutex_t  _MR_segment_mutex;

void  _MR_segment_switch(vct * vcp, int object_update);

struct _segment_stat_ {
  unsigned used;
  unsigned removed;
  unsigned compared;
  unsigned now;
  unsigned max;
}  _MR_segment_stat;

struct _dr_ {
  unsigned address;
  unsigned tid[2];
  unsigned type[2];
  unsigned segment[2];
}  _MR_dr;

/*********************************************************/

_MR_info_datarace(unsigned tid1, unsigned tid2, char * instr1, 
		  char * instr2, unsigned address,
		  vct * vc1, vct * vc2){
  FILE * file;
  char dr_file[256];

  printf("\nDatarace detected between thread %d (%s) and thread %d (%s) at address 0x%x\n",
	 tid1, instr1, tid2, instr2, address);
  
  //  _MR_address_info_static(address);
  printf("\n\n");

  //#ifdef DEBUG 
  printf("s1=%d s2=%d\n",(*vc1)[tid1],(*vc2)[tid2]);
  _MR_print_vc(vc1);
  printf("\n");
  _MR_print_vc(vc2);
  printf("\n");
  _MR_segment_statistics();
  //#endif

#ifdef STAT
  _MR_segment_statistics();
#endif

  return;

  sprintf(dr_file, "%s%s.dr", PREFIX, _jiti_naam);
  if (!(file=fopen(dr_file, "w"))) panic("info_datarace");
  fprintf(file,
	  "T%d S%d %c & T%d S%d %c @ 0x%x\n",
	  tid1, (*vc1)[tid1], instr1[0],
	  tid2, (*vc2)[tid2], instr2[0],
	  address);
  fclose(file);

  
}

/*********************************************************/

int  _MR_check_dataraces1(unsigned tid, unsigned segment_nr, 
			  bitmap bm1, bitmap bm2, char * type1, char * type2){

  int address = _MR_compare_bitmaps( bm1, bm2 );
  
  if (!address) return 0;               /* geen datarace */
  
  _MR_info_datarace(tid, 
		    _MR_segment_list[segment_nr].tid, 
		    type1, 
		    type2,
		    address,
		    (vct*) _MR_segment_now[tid].vc,
		    (vct*) _MR_segment_list[segment_nr].vc);
  return 1;
}

/*********************************************************/

int  _MR_check_dataraces(unsigned tid, unsigned segment_nr){
  int address;

   _MR_segment_stat.compared++;

  DBG(puts("+dr"));

  if (_MR_check_dataraces1(tid, segment_nr, _MR_bms[tid].bm[STORE],
			  _MR_segment_list[segment_nr].bm[LOAD], "store", "load "))
    return 1;
  
  if (_MR_check_dataraces1(tid, segment_nr, _MR_bms[tid].bm[LOAD],
			  _MR_segment_list[segment_nr].bm[STORE], "load ", "store"))
    return 1;

  if (_MR_check_dataraces1(tid, segment_nr, _MR_bms[tid].bm[STORE],
			  _MR_segment_list[segment_nr].bm[STORE], "store", "store"))
    return 1;

  DBG(puts("-dr"));
  
  return 0;
}

/*********************************************************/

_MR_update_vc(int tid, vct * object_vc, int update_type){
  int i;
  vct * t_vc;
  static pthread_mutex_t vc_mutex;

  t_vc = &(_MR_segment_now[tid].vc);

  (*t_vc)[tid]++;

  if (update_type & LOCK_VC_O) _MR_INTERN_mutex_lock(&vc_mutex);

  if (object_vc) {
    for (i=0; i<_MR_nr_cpus; i++){      
      (*t_vc)[i] = max( (*t_vc)[i], (*object_vc)[i] );
      if (update_type & UPDATE_VC_O) (*object_vc)[i]= (*t_vc)[i];      
    }      
  }

  if (update_type & LOCK_VC_O) _MR_INTERN_mutex_unlock(&vc_mutex);

}


/*********************************************************/

int _MR_segment_compare(unsigned tid){
  unsigned ta,tb;
  unsigned j;

  ta = tid;

  for(j=0; j<MAX_SEGMENTS; j++)
    if ( _MR_segment_list[j].inuse){
      tb =  _MR_segment_list[j].tid;
      
#ifdef DEBUG
      printf("CMP T%d with T%d (S%d/%d) ", ta, tb, j,  _MR_segment_list[j].invalid);
      _MR_print_vc( _MR_segment_now[ta].vc);
      _MR_print_vc( _MR_segment_list[j].vc);
#endif
      
      if ((_MR_segment_now[ta].vc[ta]>=_MR_segment_list[j].vc[ta]) && 
	  (_MR_segment_now[ta].vc[tb]<=_MR_segment_list[j].vc[tb])){
	
	if (_MR_segment_list[j].invalid) {
	  printf("\n\nPANIC: internal error (comparing with removed segment)\n\n");
	  fflush(stdout);
	  _exit(-1);  /* geen exit() gebruiken!! atexit-functies geven deadlock!! */
	} 
#ifdef DEBUG
	else
	  printf(" !! OK !!\n");
#endif	
	
	if (_MR_check_dataraces(tid, j))
	  _exit(0);
	
      }
#ifdef DEBUG
      else
	printf("\n");
#endif
      
    }
  return 0;
}

/*********************************************************/

void _MR_segment_add(int tid){
  int i;
  

  /* lege bitmaps: noch lees-, noch schrijfoperaties */
  if (!(_MR_bms[tid].bm[LOAD] || _MR_bms[tid].bm[STORE])){
    DBG(printf("--->LEGE BITMAP<---\n"));
    return; 
  }

  i=0;
  while (_MR_segment_list[i].inuse)
    i++;

  assert(i<MAX_SEGMENTS);
  
  DBG(printf("ADDed at %d\n",i));
  
  _MR_segment_list[i].inuse     = TRUE;
  _MR_segment_list[i].tid       = tid;
  _MR_segment_list[i].invalid   = FALSE;
  _MR_segment_list[i].bm[LOAD]  = _MR_bms[tid].bm[LOAD];
  _MR_segment_list[i].bm[STORE] = _MR_bms[tid].bm[STORE];
  memcpy(_MR_segment_list[i].vc, _MR_segment_now[tid].vc, sizeof(vct));
  
  _MR_bms[tid].bm[LOAD]=NULL;
  _MR_bms[tid].bm[STORE]=NULL;
  
  _MR_segment_stat.used++;
  _MR_segment_stat.now++;
  _MR_segment_stat.max=max(_MR_segment_stat.max, _MR_segment_stat.now);
  
#ifdef DEBUG  
  i=0;
  while (_MR_segment_list[i].inuse){
    printf("\n>>> %d: T%d vcp:%x bmp:%x ",i,_MR_segment_list[i].tid,_MR_segment_list[i].vc,
	   _MR_segment_list[i].bm);
    _MR_print_vc(_MR_segment_list[i].vc);
    /*
      printf("\nLEES:");      print_bitmap(segment_list[i].bm[LOAD]);
      printf("\nSCHRIJF:");   print_bitmap(segment_list[i].bm[STORE]);
      */
    i++;
  }
#endif
  
}

/*********************************************************/

void _MR_segment_remove(int tid){
  static vct vc_last;
  vct vc;
  int i,j;

  DBG(printf("\n["));
  for (i=0; i<_MR_nr_cpus; i++){
    DBG(_MR_print_vc(_MR_segment_now[i].vc));
    vc[i]=_MR_segment_now[1].vc[i];
    for (j=0; j<_MR_nr_cpus; j++){
      vc[i] =  min( vc[i], _MR_segment_now[j].vc[i] );
    }
  }
  
#ifdef DEBUG  
  printf(" => ");
  _MR_print_vc(&vc);
  printf("]\n");
#endif
  
  //  DBG(fprintf(stderr, "Bitmap size before removing obsolete segments: %db\n", _MR_bmsize));

  for(i=0; i<_MR_nr_cpus; i++)
    if(vc[i]>vc_last[i]) {
      DBG(printf(" !! DEL %d.%d -> %d.%d !!", i, vc_last[i]-1, i, vc[i]-1));
      for(j=0; j<MAX_SEGMENTS; j++)
	if ((_MR_segment_list[j].tid==i) && (_MR_segment_list[j].vc[i]<vc[i])){
	  if (!(_MR_segment_list[j].invalid)){
	    _MR_segment_list[j].invalid = TRUE; /* effectief verwijderen: ...].inuse=FALSE*/
	    _MR_segment_list[j].inuse = FALSE;
	    _MR_segment_stat.removed++;
	    _MR_segment_stat.now--;
	    _MR_remove_bitmap(_MR_segment_list[j].bm[LOAD]);
	    _MR_remove_bitmap(_MR_segment_list[j].bm[STORE]);
	  }
	}
    }
  //  DBG(fprintf(stderr, "Bitmap size after removing obsolete segments: %db\n", _MR_bmsize));

  for(i=0; i<_MR_nr_cpus; i++)
    vc_last[i]=vc[i];
 
}

/*********************************************************/

void _MR_segment_switch(vct * vcp, int update_type){
  extern int _MR_still_sequential;
  int tid=THREADID;
  
  _MR_INTERN_mutex_lock(&_MR_segment_mutex);

  DBG(printf("+----\n| STOPping segment on T%d\n| T.vc=", tid));

  _MR_segment_compare(tid);  
  _MR_segment_add(tid);
  _MR_update_vc(tid, vcp, update_type);
  //  _MR_segment_remove(tid);     

  DBG(printf("\n| STARTing segment on T%d\n+----\n", tid));
  
  DBG(_MR_segment_statistics());
  _MR_INTERN_mutex_unlock(&_MR_segment_mutex);
  
}

/*********************************************************/


_MR_print_vc(vct * vc){
  int i;

  printf("[%d]{ ", THREADID);
  for(i=1;i<_MR_nr_cpus;i++) {
    printf("%d ", (*vc)[i] );
  }
  printf("}\n");
  
}

/*********************************************************/

_MR_segment_statistics(){
  
  printf("USED: %d, REMOVED: %d, COMPARED: %d, NOW: %d, MAX: %d\n",
	 _MR_segment_stat.used,
	 _MR_segment_stat.removed,
	 _MR_segment_stat.compared,
	 _MR_segment_stat.now,
	 _MR_segment_stat.max
	 );
 
}

/*********************************************************/

int _MR_check_address(unsigned address, unsigned bytes, unsigned tid,
		      unsigned type, unsigned pc){
  static int done[2]={FALSE,FALSE};
  
  if ((!done[0]) && (tid==_MR_dr.tid[0]) && (_MR_segment_now[tid].vc[tid]==_MR_dr.segment[0])
      && (type==_MR_dr.type[0])
      && (address<=_MR_dr.address) && (_MR_dr.address<address+bytes)){
    //    _MR_address_info_dynamic(tid, address, pc, bytes, type);
    done[0]=TRUE;
    _MR_stop_thread();
  }  
  if ((!done[1]) && (tid==_MR_dr.tid[1]) && (_MR_segment_now[tid].vc[tid]==_MR_dr.segment[1]) 
      && (type==_MR_dr.type[1])
      && (address<=_MR_dr.address) && (_MR_dr.address<address+bytes)){
    //    _MR_address_info_dynamic(tid, address, pc, bytes, type);    
    done[1]=TRUE;
    _MR_stop_thread();
  }  
}

/*********************************************************/

_MR_stop_thread(){
  static int _MR_count=0;
  static pthread_mutex_t ml;

  _MR_INTERN_mutex_lock(&ml);
  _MR_count++;
  _MR_INTERN_mutex_unlock(&ml);
  
  if (_MR_count==2){
    fflush(stdout);  /* is nodig om gepijpte uitvoer mogelijk te maken */
    _exit(-1);       /* geen exit() gebruiken!! atexit-functies geven deadlock!! */
  }

}

/*******************************************************************/

int _MR_check_previous_datarace(char * prog){
  FILE * file;
  char prog_dr_file[256];
  char type0, type1;
  int n;

#if 0
  sprintf(prog_dr_file, "%s%s.dr", PREFIX, prog);

  /* indien file niet bestaat: zoek datarace */
  if (!(file=fopen(prog_dr_file, "r"))) return CHECK_DATARACES;
  
  n = fscanf(file,
	 "T%d S%d %c & T%d S%d %c @ 0x%x",
	 &_MR_dr.tid[0], &_MR_dr.segment[0], &type0,
	 &_MR_dr.tid[1], &_MR_dr.segment[1], &type1,
	 &_MR_dr.address
	 );
  fclose(file);

  /* indien foutief of lege file: enkel replay */
  if (!n) return NO_TRACING; 
  
  /* anders, reference identification */
  _MR_dr.type[0] = type0 == 'l' ? LOAD : STORE;
  _MR_dr.type[1] = type1 == 'l' ? LOAD : STORE;
  
  return REF_ID;
#else
  return 0;
#endif
}

/*******************************************************************/

int _MR_load_filter(char * prog){
  FILE * file;
  char prog_dr_file[256];
  unsigned address;

  //  _MR_filter_insert(0);  /* dummy address */

  sprintf(prog_dr_file, "%s%s.filter", PREFIX, prog);

  if (!(file=fopen(prog_dr_file, "r"))) return 0;
  
  DBG(printf("Loading filter... "));

  while (fscanf(file, "0x%x\n", &address) != EOF ){
    DBG(printf("0x%x ", address));
    // _MR_filter_insert(address);
  }

  fclose(file);

  DBG(printf("\n"));

  return 1;
}

/*******************************************************************/
