

#ifdef VIS_ONLY
#define VIS
#endif

/*
#define CHECK_LOCKING_DISCIPLINE
*/

#define NORMAL_UPDATE 0
#define SEMA_P_UPDATE 1
#define SEMA_V_UPDATE 2

#include <stdio.h>
#include "synch.h"
#include <thread.h>
#include <dlfcn.h>
#include <assert.h>
#include <pthread.h>

#include "jiti.h"
#include "segment.h"
#include "lc.h"

#ifdef RECORD
#include "record.h"
#endif

/*semaphore*/
/*
#define S_MP   (mutex_t *)&sp->pad1[0]
#define S_LCP  &((S_MP)->data.data32.LC[0])
#define S_VCPP &((S_MP)->data.data32.LC[1])
*/
#define S_MP   sp
#define S_LCP   _MR_lookup_LCp(S_MP)
#define S_VCPP  _MR_lookup_VCpp(S_MP)

/*mutex*/
/*
#define M_MP    mp
#define M_LCP   &((M_MP)->data.data32.LC[0])
#define M_VCPP  &((M_MP)->data.data32.LC[1])
#define M_OWNER *(unsigned*)(((M_MP)->data.data32.LC[1])+MAXTHREADS*4)
*/
#define M_MP    mp
#define M_LCP   _MR_lookup_LCp(M_MP)
#define M_VCPP  _MR_lookup_VCpp(M_MP)
#define M_OWNER 0

/*rwlock*/
/*
#define R_MP   (mutex_t *)(rwlp->pad1)
#define R_LCP  &((R_MP)->data.data32.LC[0])
#define R_VCPP &((R_MP)->data.data32.LC[1])
*/
#define R_MP   rwlp
#define R_LCP   _MR_lookup_LCp(R_MP)
#define R_VCPP  _MR_lookup_VCpp(R_MP)

/************************************************************************/
/*THREAD*****************************************************************/

int thr_create(void *stack_base, size_t stack_size,
	       void *(*start_func)(void *), void *arg, long flags,
	       thread_t *new_thread_ID){
  void * libs;
  int result;
  extern int _MR_still_sequential;
  static mutex_t _tidlock;
  static int (*_MR_Pthr_create)(void*, size_t,
				void *(*)(void *), void*, long,
				thread_t*)=NULL;
  
  /* voor het geval de gebruiker niet geinteresseerd is in de nieuwe tid */
  thread_t tid;
  thread_t * tidp;
  tidp = new_thread_ID ? new_thread_ID : &tid;


#ifndef RECORD
  if (_MR_still_sequential && (_MR_jiti_mode!=NO_TRACING)){
    _MR_still_sequential = FALSE;
    if (_MR_jiti_mode != REF_ID) _MR_instrumentation_on();
  }
  _MR_replay_try((LCt*)&_tidlock.LClc, 'c');
#endif


  if (!_MR_Pthr_create){   /* eerste oproep */
    if ( (libs = dlopen("/usr/lib/libthread.so",RTLD_NOW) ) == NULL)
      _MR_error(dlerror());
    if ( !(_MR_Pthr_create = (int (*)(void*, size_t,
				      void *(*)(void *), void*, long,
				      thread_t*))dlsym(libs, "thr_create")) )
      _MR_error(dlerror());
    dlclose(libs);
  //  _MR_modify_plt();  /* staat dit hier goed?? */
  /* WE MOETEN IN ELK GEVAL FLUSHEN!!! */
  }
  
#ifndef RECORD  
  if (_MR_jiti_mode!=NO_TRACING) {
    start_func = (void*(*)(void*))((unsigned)start_func + _MR_delta);
  }
#endif

  /* steeds een nieuwe LWP om deadlock te vermijden */ /* ENKEL BIJ REPLAY NODIG! */
  result = (*_MR_Pthr_create)(stack_base,
			      stack_size,
			      start_func, 
			      arg, 
			      flags | THR_SUSPENDED | THR_NEW_LWP,
			      tidp);

  if (result) _MR_error("thr_create");

  assert(*tidp<MAXTHREADS);

#ifndef RECORD  
  if (_MR_jiti_mode!=NO_TRACING) {
    _MR_segment_switch(THREADID, 0, 0);
    _MR_segment_first(*tidp);
    _MR_segment_switch(*tidp, 0, 0);
  }

  _MR_replay_next();
#else
  /*  _MR_calc_lt((LCt*)&_tidlock.LClc);*/
  _MR_calc_lt(_MR_lookup_LCp((void*)&_tidlock));
  _MR_thr_first_LC[ *tidp]=_MR_thr_LC[ *tidp ]=_MR_thr_LC[ THREADID ]+1;
#endif
  _MR_nr_threads++;


  VISONLY((LCt*)&_tidlock.LClc, THREADID, 'c');

  if (!(flags & THR_SUSPENDED))  /* thread zonder SUSPENDED vlag opgestart ? */
    thr_continue(*tidp);

  DBG(printf("T%d: ---CREATE T%d ---\n", THREADID, *tidp));



  return result;
}

/********************************************************************************/

void _MR_Ithr_exit(void *status){

  DBG(printf("T%d: ---EXIT---\n", THREADID));

#ifndef RECORD
  if (_MR_jiti_mode!=NO_TRACING) _MR_segment_last(THREADID);
#endif

  /*printf("EXIT, going to 0x%x\n", _MR_interface[THR_EXIT].R_address);*/
  (*_MR_interface[THR_EXIT].R_address)(status);

}

/********************************************************************************/

int _MR_Ithr_join(thread_t target_thread, thread_t *departed, void **status){
  int result;
  int i;
  thread_t tid;
  thread_t * tidp;
  vct * vcp;
/****
MOETEN WE DE LC HIER NIET AANPASSEN ????
****/


  tidp = departed ? departed : &tid ;
  
  result = (*_MR_interface[THR_JOIN].R_address)(target_thread, tidp, status);

#ifndef RECORD
  if (_MR_jiti_mode!=NO_TRACING) {
    vcp = &(_MR_segment_now[*tidp].vc);
    _MR_segment_switch(THREADID, &vcp, NORMAL_UPDATE);
  }
#endif

  DBG(printf("T%d: ---JOIN with %d---\n", THREADID, *tidp));
  return result;  
}

/**************************************************************************/
/*SEMA*********************************************************************/

int _MR_Isema_init(sema_t *sp, unsigned int count,
		   int type, void * arg){
  int result;
  mutex_t * mp;

  DBG(printf("T%d: ---INIT (0x%x)---\n", THREADID, sp));
  
  result =  (*_MR_interface[SEMA_INIT].R_address)(sp, count, type, arg);  
  *S_LCP=0;

  return result;
}

/**************************************************************************/

int _MR_Isema_destroy(sema_t *sp){
  int result;

  DBG(printf("T%d: ---SEMA_DESTROY (0x%x)---\n", THREADID, sp));

  result =  (*_MR_interface[SEMA_DESTROY].R_address)(sp);

  return result;
}

/**************************************************************************/

int _MR_Isema_wait(sema_t *sp){
  int result;
  
#ifndef RECORD
  _MR_replay_try(S_MP, 'w');
#endif
  
  DBG(printf("T%d: ---SEMA_WAIT (0x%x)---\n", THREADID, sp));
  
  result = (*_MR_interface[SEMA_WAIT].R_address)(sp);
  
#ifdef RECORD  
  _MR_calc_lt(S_LCP);
#else
  if (_MR_jiti_mode!=NO_TRACING)
    _MR_segment_switch(THREADID, S_VCPP, SEMA_P_UPDATE);
  _MR_replay_next();
#endif

  VISONLY(S_MP, THREADID, 'w');

  return result;
}

/**************************************************************************/

int _MR_Isema_trywait(sema_t *sp){
  int result;

#ifndef RECORD
  _MR_replay_try(S_MP, 'w');
#endif
  
  DBG(printf("T%d: ---SEMA_TRYWAIT (0x%x)---\n", THREADID, sp));
  
  result = (*_MR_interface[SEMA_TRYWAIT].R_address)(sp);

#ifdef RECORD  
  _MR_calc_lt(S_LCP);     /* !! niet atomisch !! */
#else
  if (!result)   /* sema gekregen */
    if (_MR_jiti_mode!=NO_TRACING) _MR_segment_switch(THREADID, S_VCPP, SEMA_P_UPDATE); 
  _MR_replay_next();
#endif

  VISONLY(S_MP, THREADID, 'w');

  return result;
}

/**************************************************************************/

int _MR_Isema_post(sema_t *sp){
  int result;

  VISONLY(S_MP, THREADID, 'p');

  DBG(printf("T%d: ---SEMA_POST (0x%x)---\n", THREADID, sp));

#ifdef RECORD  
  _MR_calc_lt(S_LCP);
#else 
  _MR_replay_try(S_MP, 'p');
  if (_MR_jiti_mode!=NO_TRACING)
    _MR_segment_switch(THREADID, S_VCPP, SEMA_V_UPDATE);
#endif
  
  result = (*_MR_interface[SEMA_POST].R_address)(sp);

#ifndef RECORD
  _MR_replay_next();
#endif

  return result;
}

/************************************************************************/
/*MUTEX********************************************************************/

int _MR_Imutex_init(mutex_t *mp, int type, void * arg){
  int result;

  DBG(printf("T%d: ---MUTEX_INIT (0x%x)---\n", THREADID, mp));

  result = (*_MR_interface[MUTEX_INIT].R_address)(mp, type, arg);

  *(unsigned*)M_VCPP=0;

  /*  *M_LCP=0;  */

  return result;
}

/*************************************************************************/

int _MR_Imutex_lock(mutex_t *mp){
  int result;

_MR_lookup_LCp(mp);

  
#ifndef RECORD
  _MR_replay_try(M_MP, 'l');
#endif
  
  DBG(printf("T%d: ---MUTEX_LOCK   (0x%x)---\n", THREADID, mp));

  result = (*_MR_interface[MUTEX_LOCK].R_address)(mp);

  DBG(printf("T%d: ---MUTEX_LOCK   (0x%x) done!\n", THREADID, mp));


#ifdef RECORD  
  _MR_calc_lt(M_LCP);
#else
  if (_MR_jiti_mode!=NO_TRACING){
    _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE);
  }
  /*
  M_OWNER = THREADID;
  */
  _MR_replay_next();
#endif

  VISONLY(M_MP, THREADID, 'l');

 return result;

}

/*************************************************************************/

int _MR_Imutex_trylock(mutex_t *mp){
  int result;

#ifndef RECORD
  _MR_replay_try(M_MP, 'l');
#endif
  
  DBG(printf("T%d: ---MUTEX_TRYLOCK (0x%x)---\n", THREADID, mp));

  result = (*_MR_interface[MUTEX_TRYLOCK].R_address)(mp);

#ifdef RECORD  
  _MR_calc_lt(M_LCP);     /* !! niet atomisch !! */
#else
  if (!result) { /* mutex gekregen */
    if (_MR_jiti_mode!=NO_TRACING){
      _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE); 
      /*      M_OWNER = THREADID;*/
    }
  }
  _MR_replay_next();
#endif
  
  VISONLY(M_MP, THREADID, 'l');

  return result;
}

/*************************************************************************/

int _MR_Imutex_unlock(mutex_t *mp){
  int result;

  VISONLY(M_MP, THREADID, 'u');

#ifndef RECORD
  _MR_replay_try(M_MP, 'u');
  if (_MR_jiti_mode!=NO_TRACING){
    _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE);
#ifdef CHECK_LOCKING_DISCIPLINE
    if (M_OWNER != THREADID)
      if (M_OWNER)
	_MR_error("!! mutex_unlock: mutex not unlocked by owner. Owner=t%d, Thread=%d\n",
		  M_OWNER, THREADID);
      else
	_MR_error("!! mutex_unlock: mutex not locked. Thread=%d\n", THREADID);
#endif
  }
  /*
  M_OWNER = 0;
  */
  _MR_replay_next();
#else
  _MR_calc_lt(M_LCP);
#endif
  
  DBG(printf("T%d: ---MUTEX_UNLOCK (0x%x)---\n", THREADID, mp));

  result = (*_MR_interface[MUTEX_UNLOCK].R_address)(mp);

  DBG(printf("T%d: ---MUTEX_UNLOCK (0x%x) done!\n", THREADID, mp));

  return result;
}

/************************************************************************/

int _MR_Imutex_destroy(mutex_t *mp){
  int result;

  DBG(printf("T%d: ---MUTEX_DESTROY (0x%x)---\n", THREADID, mp));
  
  result = (*_MR_interface[MUTEX_DESTROY].R_address)(mp);

  return result;
}

/************************************************************************/
/*CONDITIE VARIABELEN****************************************************/

int _MR_Icond_init(cond_t *cvp, int type, int arg){
  int result;

  DBG(printf("T%d: ---COND_INIT (0x%x)---\n", THREADID, cvp));
  
  result = (*_MR_interface[COND_INIT].R_address)(cvp, type, arg);

  return result;
}

/************************************************************************/

int _MR_Icond_wait(cond_t *cvp, mutex_t *mp){
  int result;
  
  DBG(printf("T%d: ---COND_WAIT (C0x%x, M0x%x)---\n", THREADID, cvp, mp));
  
  
#ifndef RECORD
  if (_MR_replay_mode) {
    (*_MR_interface[MUTEX_UNLOCK].R_address)(mp);
    _MR_replay_try(M_MP, '?');
    (*_MR_interface[MUTEX_LOCK].R_address)(mp);
  }

  if (_MR_jiti_mode!=NO_TRACING){
    _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE);
#ifdef CHECK_LOCKING_DISCIPLINE
    if (M_OWNER != THREADID)
      if (M_OWNER)
	_MR_error("!! cond_wait: mutex not unlocked by owner. Owner=t%d, Thread=%d\n",
		  M_OWNER, THREADID);
      else
	_MR_error("!! cond_wait: mutex not locked. Thread=%d\n", THREADID);
#endif
  }

  _MR_replay_next();
#else
  
  /*  _MR_calc_lt(M_LCP);*/
  
#endif
  
  
#ifdef RECORD
/*** MOET HIER OOK GEEN _MR-calc_lt STAAN ???? ****/
  result = (*_MR_interface[COND_WAIT].R_address)(cvp, mp);
  _MR_calc_lt(M_LCP);
#else
  /*  if (_MR_jiti_mode!=NO_TRACING) M_OWNER=0;*/
  
  if (!_MR_replay_mode) result = (*_MR_interface[COND_WAIT].R_address)(cvp, mp);
  
  /* _MR_replay_try(M_MP);*/
  if (_MR_jiti_mode!=NO_TRACING){
    _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE);
    
    /*    M_OWNER = THREADID;*/
  }
  /*_MR_replay_next();*/
  
#endif
  
  return result;
  
}

/************************************************************************/

int _MR_Icond_timedwait(cond_t *cvp, mutex_t *mp,
			timestruc_t *abstime){
  int result;
  
  DBG(printf("T%d: ---COND_TIMEDWAIT (0x%x)---\n", THREADID, cvp, mp));

  /* unlock */
#ifndef RECORD
  _MR_replay_try(M_MP, '?');
  if (_MR_jiti_mode!=NO_TRACING){
    _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE);
#ifdef CHECK_LOCKING_DISCIPLINE
    if (M_OWNER != THREADID)
      if (M_OWNER)
	_MR_error("!! cond_timedwait: mutex not unlocked by owner. Owner=t%d, Thread=%d\n",
		  M_OWNER, THREADID);
      else
	_MR_error("!! cond_timedwait: mutex not locked. Thread=%d\n", THREADID);
#endif
  }
  _MR_replay_next();
#else
  _MR_calc_lt(M_LCP);
#endif
  
  /* nog aanpassen */
  result = (*_MR_interface[COND_TIMEDWAIT].R_address)(cvp, mp, abstime);
  
  /* lock */
#ifdef RECORD  
  _MR_calc_lt(M_LCP);
#else
  _MR_replay_try(M_MP, '?');
  if (_MR_jiti_mode!=NO_TRACING){
    _MR_segment_switch(THREADID, M_VCPP, NORMAL_UPDATE);
    /*    M_OWNER = THREADID;*/
  }
  _MR_replay_next();
#endif
  
  return result;
}

/************************************************************************/

int _MR_Icond_signal(cond_t *cvp){
  int result;

  DBG(printf("T%d: ---COND_SIGNAL (0x%x)---\n", THREADID, cvp));

#ifdef RECORD  
  result = (*_MR_interface[COND_SIGNAL].R_address)(cvp);
#else
  if (!_MR_replay_mode) result = (*_MR_interface[COND_SIGNAL].R_address)(cvp);
#endif
  
  return result;
}

/************************************************************************/

int _MR_Icond_broadcast(cond_t *cvp){
  int result;

  DBG(printf("T%d: ---COND_BROADCAST (0x%x)---\n", THREADID, cvp));

#ifdef RECORD
  result = (*_MR_interface[COND_BROADCAST].R_address)(cvp);
#else
  if (!_MR_replay_mode) result = (*_MR_interface[COND_BROADCAST].R_address)(cvp);
#endif

  return result;
}

/************************************************************************/

int _MR_Icond_destroy(cond_t *cvp){
  int result;

  DBG(printf("T%d: ---COND_DESTROY (0x%x)---\n", THREADID, cvp));
  
  result = (*_MR_interface[COND_DESTROY].R_address)(cvp);

  return result;
}

/************************************************************************/
/*RWLOCK*****************************************************************/

int _MR_Irwlock_init(rwlock_t *rwlp, int type, void * arg){
  int result;

  DBG(printf("T%d: ---RWLOCK_INIT (0x%x)---\n", THREADID, rwlp));
  
  result = (*_MR_interface[RWLOCK_INIT].R_address)(rwlp, type, arg);

  return result;
}

/************************************************************************/

int _MR_Irwlock_destroy(rwlock_t *rwlp){
  int result;

  DBG(printf("T%d: ---RWLOCK_DESTROY (0x%x)---\n", THREADID, rwlp));
  
  result = (*_MR_interface[RWLOCK_DESTROY].R_address)(rwlp);

  return result;
}

/************************************************************************/

int _MR_Irw_rdlock(rwlock_t *rwlp){
  int result;

#ifndef RECORD
  _MR_replay_try(R_MP, '?');
#endif

  DBG(printf("T%d: ---RW_RDLOCK (0x%x)---\n", THREADID, rwlp));
  
  result = (*_MR_interface[RW_RDLOCK].R_address)(rwlp);

#ifdef RECORD  
  _MR_calc_lt(R_LCP);
#else
  if (_MR_jiti_mode!=NO_TRACING)
    _MR_segment_switch(THREADID, R_VCPP, SEMA_P_UPDATE);
  _MR_replay_next();
#endif

  return result;
}

/************************************************************************/

int _MR_Irw_wrlock(rwlock_t *rwlp){
  int result;

#ifndef RECORD
  _MR_replay_try(R_MP, '?');
#endif

  DBG(printf("T%d: ---RW_WRLOCK (0x%x)---\n", THREADID, rwlp));
  
  result = (*_MR_interface[RW_WRLOCK].R_address)(rwlp);

#ifdef RECORD
  _MR_calc_lt(R_LCP);
#else
  if (_MR_jiti_mode!=NO_TRACING)
    _MR_segment_switch(THREADID, R_VCPP, NORMAL_UPDATE);
  _MR_replay_next();
#endif

  return result;
}

/************************************************************************/

int _MR_Irw_unlock(rwlock_t *rwlp){
  int result;

#ifndef RECORD
  _MR_replay_try(R_MP, '?');

  if (rwlp->readers>0)  /* unlock een read-lock */
    if (_MR_jiti_mode!=NO_TRACING)
      _MR_segment_switch(THREADID, R_VCPP, SEMA_P_UPDATE);
    else                /* unlock een write-lock */
      if (_MR_jiti_mode!=NO_TRACING)
	_MR_segment_switch(THREADID, R_VCPP, NORMAL_UPDATE);
  /*  M_OWNER = 0; */ 
  _MR_replay_next();
#else
  _MR_calc_lt(R_LCP);
#endif
  
  DBG(printf("T%d: ---RW_UNLOCK (0x%x)---\n", THREADID, rwlp));

  result = (*_MR_interface[RW_UNLOCK].R_address)(rwlp);

  return result;
}

/************************************************************************/

int _MR_Irw_tryrdlock(rwlock_t *rwlp){
  int result;

#ifndef RECORD
  _MR_replay_try(R_MP, '?');
#endif

  DBG(printf("T%d: ---RW_TRYRDLOCK (0x%x)---\n", THREADID, rwlp));
  
  result = (*_MR_interface[RW_TRYRDLOCK].R_address)(rwlp);

#ifdef RECORD  
  _MR_calc_lt(R_LCP);
#else
  if (!result)
    if (_MR_jiti_mode!=NO_TRACING)
      _MR_segment_switch(THREADID, R_VCPP, SEMA_P_UPDATE); /* lock gekregen */
  _MR_replay_next();
#endif
  
  return result;
}

/************************************************************************/

int _MR_Irw_trywrlock(rwlock_t *rwlp){
  int result;

#ifndef RECORD
  _MR_replay_try(R_MP, '?');
#endif

  DBG(printf("T%d: ---RW_TRYWRLOCK (0x%x)---\n", THREADID, rwlp));
  
result = (*_MR_interface[RW_TRYWRLOCK].R_address)(rwlp);

#ifdef RECORD  
  _MR_calc_lt(R_LCP);
#else
  if (!result) 
    if (_MR_jiti_mode!=NO_TRACING)
      _MR_segment_switch(THREADID, R_VCPP, NORMAL_UPDATE); /* lock gekregen */
  _MR_replay_next();
#endif

  return result;
}

