//#define DEBUG

#include <limits.h>
#include <stdio.h>
#include <dlfcn.h>
#include <assert.h>

#include "jiti-recplay.h"


LCt _MR_thr_LC[MAXTHREADS];
LCt _MR_thr_LOGS[ MAXTHREADS ][ LOGSPACE];
unsigned int _MR_thr_LOG_i[MAXTHREADS];
unsigned int _MR_thr_LOG_i_max[MAXTHREADS];
unsigned int _MR_thr_last_LC[MAXTHREADS];

unsigned int _MR_thr_REPLAY_blocks[MAXTHREADS];

#include <semaphore.h>

sem_t _MR_deze_semafoor_wordt_nooit_vrijgegeven;

pthread_mutex_t _MR_file_lock, _MR_replay_mutex;

typedef FILE * fp;
fp _MR_fp_t;

pthread_cond_t _MR_cond2[MAXTHREADS];


/*************************************************************************/

LCt _MR_nextLC(unsigned int tid);
void _MR_loadlog(unsigned int tid);

/*************************************************************************/

int _MR_replay_try(){
  int i, OK;
  int tid=THREADID;
  LCt LC;

  //  if ((tid==1) && (_MR_thr_LC[1]==0)) _MR_thr_LC[1]=1;

  LC = _MR_thr_LC[tid];
  
  if (!LC){
    DBG(printf("WARNING: T%d died without using thr_exit()!\n", THREADID));
    _MR_INTERN_sem_wait(&_MR_deze_semafoor_wordt_nooit_vrijgegeven);
  }
  
  _MR_INTERN_mutex_lock(&_MR_replay_mutex); /*---------------*/
  
  DBG(printf("\nT%d checking...\n", THREADID));

  OK=FALSE;
  while ( !OK ){
    OK=TRUE;
    i=0;
    while (++i < MAXTHREADS ) {
      DBG(printf("[%d:%d]",i,_MR_thr_LC[i]));
      if ((_MR_thr_LC[i] < LC) && (_MR_thr_LC[i])){
	OK=FALSE;
      }
    }
    DBG(printf("\n"));
    if (!OK) {
      DBG(printf("T%d sleeping...\n", THREADID));
      _MR_INTERN_cond_wait(&_MR_cond2[THREADID],&_MR_replay_mutex);
      DBG(printf("\nT%d: Trying again...\n", THREADID));
    }
  }
  
  DBG(printf("T%d succeeded!\n",THREADID));

  _MR_INTERN_mutex_unlock(&_MR_replay_mutex); /*---------------*/
  
  return OK;
}

/*************************************************************************/

int _MR_replay_next(){
  unsigned int tid = THREADID;
  unsigned int i;
  unsigned int min;

  _MR_thr_LC[tid] = _MR_nextLC(tid);
  min=UINT_MAX;

  _MR_INTERN_mutex_lock(&_MR_replay_mutex); /*---------------*/

  for (i=0; i<MAXTHREADS; i++)
    if ((_MR_thr_LC[i]) && (_MR_thr_LC[i]<min))
      min=_MR_thr_LC[i];
  DBG(printf("\nT%d: seeking next LC: MIN=%u: signaling ", THREADID, min));
  for (i=0; i<MAXTHREADS; i++)
    if ((i!=THREADID) && (min==_MR_thr_LC[i])){
      DBG(printf("%u ", i));
      _MR_INTERN_cond_signal(&_MR_cond2[i]);
      DBG(printf("SIGNAL from T%d (new LC=%d) to T%d\n",THREADID,_MR_thr_LC[tid],i));
    }
  DBG(printf("\n"));
  _MR_INTERN_mutex_unlock(&_MR_replay_mutex); /*---------------*/
  
}

/*************************************************************************/

LCt _MR_nextLC(unsigned int tid){
  int pointer;

  DBG(printf("tid=%d, _MR_thr_LOG_i=%d, _MR_thr_LOG_i_max=%d, _MR_thr_LC=%d, _MR_thr_last_LC=%d\n",
	     tid, _MR_thr_LOG_i[tid], _MR_thr_LOG_i_max[tid], _MR_thr_LC[tid], _MR_thr_last_LC[tid]));
  
  /* thread heeft geen objecten meer nodig, of heeft er nooit gebruikt */
  if ((_MR_thr_last_LC[tid]==0) ||  (_MR_thr_LC[tid] >= _MR_thr_last_LC[tid])) return 0;   

  pointer = _MR_thr_LOG_i[ tid ];

  if (pointer > _MR_thr_LOG_i_max[ tid ]) {
    /*nieuw blok laden*/
    _MR_loadlog(tid);
    pointer=0;
  }
  
  if ( !_MR_thr_LOG_i_max[ tid ]){              /* Bestand is opgebruikt, geen nieuw blok meer */
    
    _MR_thr_LC[ tid ]++;
    
  } else {  
    
    int first = (LCt)_MR_thr_LOGS[ tid ][ pointer ];
    int last  = (LCt)_MR_thr_LOGS[ tid ][ pointer+1 ];

    DBG(printf("first=%d, last=%d\n", first, last));

    if ( _MR_thr_LC[ tid ] == first ){
      _MR_thr_LC[ tid ]     = last;
      _MR_thr_LOG_i[ tid ] += 2;
    } else
      _MR_thr_LC[ tid ]++;    
  }

  return _MR_thr_LC[tid];
  
}

/*************************************************************************/

void _MR_loadlog(unsigned int tid){
  unsigned int i,w,block, OK, F_EOF, block_tid, block_count, block_size;
 
#ifdef INCREMENT
  unsigned buffer[LOGSPACE*5/4];
  unsigned buffer_index;
  unsigned key,a,b,lastb;
  static unsigned int _MR_thr_LOG_lastb[MAXTHREADS];
#endif  

#define CHECK_EOF        \
    if (feof(_MR_fp_t)){ \
      OK = TRUE;         \
      F_EOF = TRUE;      \
    }

  block = ++_MR_thr_REPLAY_blocks[tid];
  block_count=0;

  _MR_INTERN_mutex_lock(&_MR_file_lock);
  
  rewind(_MR_fp_t);
  OK = FALSE;
  F_EOF = FALSE;

  while (!OK){
    block_tid = getw(_MR_fp_t);    CHECK_EOF;

    /* Read blocks until we have a block with the correct thread identifier */
    while ((block_tid != tid) && (!F_EOF)) {
      block_size = getw(_MR_fp_t);      
      fseek(_MR_fp_t, block_size*sizeof(char), SEEK_CUR);
      block_tid = getw(_MR_fp_t);      CHECK_EOF;      
    }

    if (F_EOF) goto file_done;

    /* Check if the number of the block corresponds to the block we need */
    block_count++;
    if (block_count == block)
      OK=TRUE;
    else {
      OK=FALSE;
      block_size = getw(_MR_fp_t);
      fseek(_MR_fp_t, block_size*sizeof(char), SEEK_CUR);
    }

  }
  
  if (!F_EOF) {                      /* blok bestaat */
    block_size=getw(_MR_fp_t);

#ifdef INCREMENT
    if (tid) {
      fread(buffer, sizeof(char), block_size, _MR_fp_t);
      
      buffer_index = (unsigned)buffer;
      i=0;
      
      lastb=_MR_thr_LOG_lastb[ tid ];
      
      while ((buffer_index-(unsigned)buffer) < block_size) {
	
	key = *(unsigned char*)buffer_index++;
	if (key!=255)
	  a = (unsigned)key;
	else {
	  a = ((*(unsigned char*)buffer_index)<<24)+
	    ((*(unsigned char*)(buffer_index+1))<<16)+
	    ((*(unsigned char*)(buffer_index+2))<<8)+
	    (*(unsigned char*)(buffer_index+3));
	  buffer_index+=4;
	}
	
	key = *(unsigned char*)buffer_index++;
	if (key!=255)
	  b = (unsigned)key;
	else {
	  b = ((*(unsigned char*)buffer_index)<<24)+
	    ((*(unsigned char*)(buffer_index+1))<<16)+
	    ((*(unsigned char*)(buffer_index+2))<<8)+
	    (*(unsigned char*)(buffer_index+3));
	  buffer_index+=4;
	}
	
	a = a + lastb;
	b = b + a + 2;
	lastb = b;
	
	_MR_thr_LOGS[ tid ][ i++ ] = a;
	_MR_thr_LOGS[ tid ][ i++ ] = b;
      } 
      _MR_thr_LOG_lastb[ tid ]=lastb;
      block_size = i-1;
      _MR_thr_LOG_i_max[ tid ] = block_size;
    } else {   // tid=0 => geen incrementen
      _MR_thr_LOG_i_max[ tid ] = block_size/4 - 1;
      fread(&_MR_thr_LOGS[ tid ], sizeof(char), block_size, _MR_fp_t);
    }
#else
    _MR_thr_LOG_i_max[ tid ] = block_size/4 - 1;
    fread(&_MR_thr_LOGS[ tid ], sizeof(char), block_size, _MR_fp_t);
#endif
    _MR_thr_LOG_i[ tid ] = 0;

  } else {                         /* blok bestaat niet */
  file_done:
    _MR_thr_LOG_i[ tid ] = 0;
    _MR_thr_LOG_i_max[ tid ] = 0;
  }

  _MR_INTERN_mutex_unlock(&_MR_file_lock); 

}

/*************************************************************************/

int _MR_start_replay(){
  int i;
  char naam[100];
  char * _MR_prog = jiti_program_name();
 
  DBG(printf("-> Initializing replay facility\n"));

  sprintf(naam,"%s%s.%c", PREFIX, _MR_prog, 't'); 
  cleanup_file_name(naam);

  if (!(_MR_fp_t=fopen(naam,"r"))) return FALSE;
  for (i=0; i<MAXTHREADS ; i++){
    _MR_thr_LOG_i[ i ] = 2;  // forceer i > i_max  => dit veroorzaakt een loadlog() bij eerste gebruik
    _MR_thr_LOG_i_max[ i ] = 1;  // 
  }

  // lees trace van thread 0, deze bevat de 'f' en 'l' informatie
  _MR_loadlog(0);
  for (i=0; i<MAXTHREADS; i++){
    _MR_thr_LC[i]      = _MR_thr_LOGS[0][i];             // first
    _MR_thr_last_LC[i] = _MR_thr_LOGS[0][MAXTHREADS+i];  // last
    _MR_thr_LC[i]      = _MR_nextLC(i);                  // --> next, hier worden loadlogs geforceerd
                                                         //     dit moet dus NA het bepalen van  
                                                         //     _MR_thr_last_LC gebeuren !!!
  }

  //// Force a jumpstart for the first thread.
  _MR_thr_LC[1]=2;

  return TRUE;
}

/************************************************************************/

int _MR_stop_replay(){
  close(_MR_fp_t);
}


