/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  bandwith.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/11/12 10:13:26 $
 * $Log: bandwidth.c,v $
 * Revision 1.4  2001/11/12 10:13:26  ronsse
 * Added comments
 *
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend calculates the number of bytes read/written 
 * as code and as data.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"

static unsigned _bandwidth_data=0;
static unsigned _bandwidth_code=0;

/////////////////////////////////////

// Our registered callback for data accesses. We receive the instruction pointer, the
// instruction type (load, store, change), the number of bytes accesses and the memory
// address used.
static _bandwidth_data_trace(unsigned ip, unsigned type, unsigned len, unsigned address){  
  _bandwidth_data += ( type==MODIFY ? 2 : 1 ) * len;
}

/////////////////////////////////////

// Our registered callback for code accesses. Code is executed between instructions
// at addresses 'from' and 'to'.
static _bandwidth_code_trace(unsigned from, unsigned to){
  _bandwidth_code += to-from+1;
}

//////////////////////////////////////////////////

// This function is called as soon as the program ends.
static _bandwidth_done(){
  fprintf(jiti_tty, "===== Bandwidth: done\n");
  fprintf(jiti_tty, "Code: %d bytes (%f%%)\n", _bandwidth_code, _bandwidth_code*100.0/(_bandwidth_code+_bandwidth_data));
  fprintf(jiti_tty, "Data: %d bytes (%f%%)\n", _bandwidth_data, _bandwidth_data*100.0/(_bandwidth_code+_bandwidth_data));
}

//////////////////////////////////////////////////

// Our initialization routine
static _bandwidth_init(){

  fprintf(jiti_tty, "===== Bandwidth ======\n");

  // Tell JiTI86 that we want to know all memory operations, the code execution, and the end of the
  // program. The callbacks registered will be called by JiTI86, each time an interesting event
  // occurs.
  
  jiti_callback("#DATA#", (unsigned)_bandwidth_data_trace);
  jiti_callback("#CODE#", (unsigned)_bandwidth_code_trace);
  jiti_callback("#STOP#", (unsigned)_bandwidth_done);

}

/////////////////////////////////////

// This notifies JiTI86 that _bandwidth_init our initialization routine is

INIT(_bandwidth_init);                     

