/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti-mem.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/10/25 13:46:01 $
 * $Log: trace-mem.c,v $
 * Revision 1.2  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend generates a list of memory accesses perfromed
 * during a program run. For each memory access, the following infromation is shown:
 * 	-the type of operation: S(tore), L(oad) or X(=load+store)
 * 	-the identifier of the thread perfroming the operation (for parallel programs)
 * 	-the address of the instruction performing the memory operation
 * 	-the memory address used
 * 	-the number of bytes read/written
 * 	-the assembly instruction
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"

char c_type[3]={'L', 'S', 'X'};

// uitvoerformaat voor het data-adres
#define OUTPUT_FORMAT "%.8x"    // hexadeximaal
#define _OUTPUT_FORMAT "%u"      // decimaal

/////////////////////////////////////

static void trace_data_seq(unsigned ip, unsigned type, unsigned len, unsigned address){
  
  fprintf(jiti_tty, "%c 1 %.8x: " OUTPUT_FORMAT " %4d ", c_type[type], ip, address, len);
  jiti_fprint_ins1(jiti_tty, 0, ip);
  fprintf(jiti_tty, "\n");

}

/////////////////////////////////////

static void trace_data_par(unsigned ip, unsigned type, unsigned len, unsigned address){
  unsigned p=THREADID;
  static pthread_mutex_t mutex;
  
  pthread_mutex_lock(&mutex);

  fprintf(jiti_tty, "%c %d %.8x: " OUTPUT_FORMAT " %4d ", c_type[type], p, ip, address, len);
  jiti_fprint_ins1(jiti_tty, 0, ip);
  fprintf(jiti_tty, "\n");
  
  pthread_mutex_unlock(&mutex);
}

/////////////////////////////////////

static init_trace_data(){

  fprintf(jiti_tty, "===== MEM_INIT: Trace all data accesses ======\n");

  jiti_callback("#DATA#", (unsigned)(jiti_parallel_program() ? trace_data_par : trace_data_seq));

}

/////////////////////////////////////

INIT(init_trace_data);

