/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  callback.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/10/25 13:46:01 $
 * $Log: callback.c,v $
 * Revision 1.4  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"

typedef struct _callback_t_ {
    char *function_name;
    void (*function) (unsigned, unsigned, unsigned, unsigned);	// MAX 4 ARGUMENTEN!!
    struct _callback_t_ *next;
} callback_t;

#define MAX_CALLBACKS 4		// START, STOP, DATA, CODE

callback_t callbacks[MAX_CALLBACKS];

/////////////////////////////////////////////////////////////////////////////

void execute_callbacks(unsigned callback, unsigned arg1, unsigned arg2, unsigned arg3, unsigned arg4){
    callback_t *loper = &(callbacks[callback]);
    DBG(fprintf(jiti_tty, "\n>>> Executing callbacks for %d", callback));
    while (loper->next) {
	(*loper->function) (arg1, arg2, arg3, arg4);
	loper = loper->next;
    }
    (*loper->function) (arg1, arg2, arg3, arg4);
}

#define make_callback_executor(XXX);\
	void execute_callback_##XXX(unsigned arg1, unsigned arg2, unsigned arg3, unsigned arg4){\
		execute_callbacks(XXX, arg1, arg2, arg3, arg4);\
	}

make_callback_executor(0);
make_callback_executor(1);
make_callback_executor(2);
make_callback_executor(3);

void (*callback_addresses[MAX_CALLBACKS]) (unsigned, unsigned, unsigned, unsigned)={
    execute_callback_0, 
    execute_callback_1, 
    execute_callback_2,
    execute_callback_3
};

/////////////////////////////////////////////////////////////////////////////

extern unsigned CB_stop;
extern unsigned CB_start;
extern unsigned CB_data;
extern unsigned CB_code;

void link_jiti_with_callbacks(char *name, unsigned callback) {

    DBG(fprintf(jiti_tty, "\n>>> Linking %s with %d-th callback", name, callback));

    if (!strcmp(name, "#STOP#")) {

	CB_stop = (unsigned) callback_addresses[callback];

    } else if (!strcmp(name, "#START#")) {

	CB_start = (unsigned) callback_addresses[callback];

    } else if (!strcmp(name, "#DATA#")) {

	CB_data = (unsigned) callback_addresses[callback];
	_instrument_mem_ = 1;

    } else if (!strcmp(name, "#CODE#")) {

	CB_code = (unsigned) callback_addresses[callback];
	_instrument_code_ = 1;

    } else {

	panic("link_jiti_with_callbacks");

    }

}

//////////////////////////////////////////////////////////////////////////////

void __insert_callback(char *name, void (*function)(unsigned, unsigned, unsigned, unsigned)){
    int i = 0;

    while (i < MAX_CALLBACKS) {

	if (callbacks[i].function_name == NULL) {

	    // lege plaats gevonden
	    DBG(fprintf(jiti_tty, "empty place @ %d\n", i));
	    strcpy(callbacks[i].function_name =
		   malloc(strlen(name)), name);
	    callbacks[i].function = function;
	    callbacks[i].next = NULL;

	    // nu nog zorgen dat JiTI de callback oproept:
	    link_jiti_with_callbacks(name, i);
	    return;

	} else {

	    if (!strcmp(callbacks[i].function_name, name)) {
		callback_t *last_callback = &(callbacks[i]);

		// functie gevonden, voeg achteraan toe!!
		DBG(fprintf(jiti_tty, "found @ %d\n", i));

		while (last_callback->next)
		    last_callback = last_callback->next;

		last_callback->next = malloc(sizeof(callback_t));
		last_callback = last_callback->next;
		last_callback->function_name = NULL;	// not used
		last_callback->function = function;
		last_callback->next = NULL;

		return;

	    } else {
		// probeer volgende
		i++;
	    }

	}
    }				// while

    // te weinig plaats!
    panic("callbacks[] too small!");

}

/////////////////////////////////////////////////////////////////

void ___internal___error___001(){
    panic("___internal___error___#001");
}

/////////////////////////////////////////////////////////////////

unsigned jiti_callback(char *function, unsigned callback){

    if (function[0] == '#') {	// special case

	__insert_callback(function, (void(*)(unsigned, unsigned, unsigned, unsigned))callback);
	return (unsigned) ___internal___error___001;

    } else {			// dyn. linked function

	_instrument_dyn_ = 1;

	/* 
	 * PAS OP: teruggave van NULL is mogelijk indien intercept_plt() niets in de PLT-tabel vindt. Dit treedt
	 * op indien de applikatie de desbetreffende functie niet gebruikt. NULL teruggeven is dan OK,
	 * de functie wordt toch nooit opgeroepen!
	 */
	return intercept_plt((char *) function, callback);;

    }
}
