
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  elf.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/11/12 10:06:47 $
 * $Log: elf.c,v $
 * Revision 1.4  2001/11/12 10:06:47  ronsse
 * Inlezen van PLT gaat nu correct.
 *
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>   
#include <elf.h>
#include <unistd.h>    
#include <stdlib.h>

#include "jiti.h"

/* process_elf_info vult volgende addressen in: */
extern unsigned _elf_start;
extern unsigned _elf_init;
extern unsigned _elf_fini;

/* en bouwt de PLT-tabel op */

void process_elf_info(char * executable){
  char buffer[sizeof(Elf32_Ehdr)];
  char buffer2[1000];
  Elf32_Ehdr * elfheader;
  Elf32_Phdr * elfprogram;
  Elf32_Shdr * elfsection;
  int i;
  char * dyn_info=NULL;
  unsigned dyn_size=0;
  unsigned plt_got_address=0, sym_tab=0, str_tab=0, sym_tab_num=0, str_tab_size=0;
  
  int file=open(executable, O_RDONLY);

  // elf header
  read(file, buffer, sizeof(Elf32_Ehdr));
  elfheader=(Elf32_Ehdr *)buffer;
  _elf_start=elfheader->e_entry;
  
  // program segment headers
  for (i=0; i<elfheader->e_phnum; i++) {
    lseek(file, elfheader->e_phoff+i*elfheader->e_phentsize, SEEK_SET);
    read(file, buffer2, elfheader->e_phentsize);
    elfprogram=(Elf32_Phdr *)buffer2;
    if (elfprogram->p_type==PT_DYNAMIC) {
     dyn_info = (char*)elfprogram->p_offset;
     dyn_size = (unsigned)elfprogram->p_filesz;
     break;
    }
  }

  sym_tab_num=0;
  for (i=0; i<elfheader->e_shnum; i++) {
    lseek(file, elfheader->e_shoff+i*elfheader->e_shentsize, SEEK_SET);
    read(file, buffer2, elfheader->e_shentsize);
    elfsection=(Elf32_Shdr *)buffer2;
    if (elfsection->sh_type==SHT_DYNSYM)
    {
      sym_tab_num=elfsection->sh_size;
      break;
    }
  }

  if (sym_tab_num==0)
  {
    printf("no dynamic symtab\n"); exit(0);
  }

  // dyn link info
  {
    Elf32_Dyn * dyn_entry;
    unsigned * buffer3=(unsigned*)malloc(dyn_size);

    lseek(file, (unsigned)dyn_info, SEEK_SET);
    read(file, buffer3, dyn_size);    
    dyn_entry=(Elf32_Dyn*)buffer3;

    while (dyn_entry->d_tag!=DT_NULL){
      switch(dyn_entry->d_tag){
      case DT_INIT:
	_elf_init=dyn_entry->d_un.d_val;
	DBG(fprintf(jiti_tty, "\n>>> INIT @ 0x%x", dyn_entry->d_un.d_val));
	break;
      case DT_FINI:
	_elf_fini=dyn_entry->d_un.d_val;
	DBG(fprintf(jiti_tty, "\n>>> FINI @ 0x%x", dyn_entry->d_un.d_val));
	break;
      case DT_PLTGOT:
	DBG(fprintf(jiti_tty, "\n>>> GOT @ 0x%x", dyn_entry->d_un.d_val));
	plt_got_address = dyn_entry->d_un.d_val+12;
	break;
      case DT_PLTRELSZ:
	//plt_got_size = dyn_entry->d_un.d_val/2;
	break;
      case DT_SYMTAB:
	DBG(fprintf(jiti_tty, "\n>>> Symbol table @ 0x%x", dyn_entry->d_un.d_val));
	sym_tab=dyn_entry->d_un.d_val;
	break;
      case DT_SYMENT:
	DBG(fprintf(jiti_tty, "\n>>> Symbol table size= %d", dyn_entry->d_un.d_val));
	sym_tab_num/=dyn_entry->d_un.d_val;
	break;
      case DT_STRSZ:
	DBG(fprintf(jiti_tty, "\n>>> String table size= %d", dyn_entry->d_un.d_val));
	str_tab_size=dyn_entry->d_un.d_val;
	break;
      case DT_STRTAB:
	DBG(fprintf(jiti_tty, "\n>>> String table @ 0x%x", dyn_entry->d_un.d_val));
	str_tab=dyn_entry->d_un.d_val;
	break;
      }
      dyn_entry++;
    }
  }
  
  close(file);

  // sym table
  {
    Elf32_Sym * symbol = (Elf32_Sym*)sym_tab;
    unsigned jmp_address /*, real_target*/;
    int i=0;

    while (i<sym_tab_num) {

	    // Black magic
//	    if (symbol->st_info==0x74) 
//		    break;

	    DBG(fprintf(jiti_tty, "\n>>> [%2d] Found %-25s: type=%x, addres=%x ", 
				    i,
				    str_tab+symbol->st_name, 
				    symbol->st_info,
				    symbol->st_value
		       ));

      if (ELF32_ST_TYPE(symbol->st_info)==STT_FUNC) {
	jmp_address=*(unsigned*)(symbol->st_value+2);
	DBG(fprintf(jiti_tty, " === Jump addres=%x", jmp_address));
	add_to_plt((char*)str_tab+symbol->st_name, jmp_address);
      }

      symbol++;
      i++;
    }    
  }  
}



