/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  fp.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/10/25 13:46:01 $
 * $Log: fp.c,v $
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"


extern unsigned char byte1, byte2, byte3, byte4, prefix_c, immsize_override;

struct float_len_type {
  char len;
  char type;
} float_len_type[64] = {
  /*  0 */ {  4, 0 },  /* FADD */
  /*  1 */ {  4, 0 },  /* */
  /*  2 */ {  4, 0 },  /* FCOM */
  /*  3 */ {  4, 0 },  /* */
  /*  4 */ {  4, 0 },  /* */
  /*  5 */ {  4, 0 },  /* */
  /*  6 */ {  4, 0 },  /* FDIV */
  /*  7 */ {  4, 0 },  /* FDIVR */
  /*  8 */ {  4, 0 },  /* */
  /*  9 */ {  0, 0 },  /* */
  /* 10 */ {  4, 1 },  /* */
  /* 11 */ {  4, 1 },  /* */
  /* 12 */ { 28, 0 },  /* */
  /* 13 */ {  2, 0 },  /* */
  /* 14 */ { 28, 1 },  /* */
  /* 15 */ {  2, 1 },  /* */
  /* 16 */ {  4, 0 },  /* FIADD */
  /* 17 */ {  4, 0 },  /* */
  /* 18 */ {  4, 0 },  /* FICOM */
  /* 19 */ {  4, 0 },  /* FICOMP */
  /* 20 */ {  4, 0 },  /* */
  /* 21 */ {  4, 0 },  /* */
  /* 22 */ {  4, 0 },  /* */
  /* 23 */ {  4, 0 },  /* */
  /* 24 */ {  4, 0 },  /* */
  /* 25 */ {  0, 0 },  /* */
  /* 26 */ {  4, 1 },  /* */
  /* 27 */ {  4, 1 },  /* */
  /* 28 */ {  0, 0 },  /* */
  /* 29 */ { 10, 0 },  /* */
  /* 30 */ {  0, 0 },  /* */
  /* 31 */ { 10, 1 },  /* */
  /* 32 */ {  8, 0 },  /* FADD */
  /* 33 */ {  8, 0 },  /* */
  /* 34 */ {  8, 0 },  /* FCOM */
  /* 35 */ {  8, 0 },  /* */
  /* 36 */ {  8, 0 },  /* */
  /* 37 */ {  8, 0 },  /* */
  /* 38 */ {  8, 0 },  /* FDIV */
  /* 39 */ {  8, 0 },  /* FDIVR */
  /* 40 */ {  8, 0 },  /* */
  /* 41 */ {  0, 0 },  /* */
  /* 42 */ {  8, 1 },  /* */
  /* 43 */ {  8, 1 },  /* */
  /* 44 */ {108, 0 },  /* */
  /* 45 */ {  0, 0 },  /* */
  /* 46 */ {108, 1 },  /* */
  /* 47 */ {  2, 0 },  /* */
  /* 48 */ {  2, 0 },  /* FIADD */
  /* 49 */ {  2, 0 },  /* */
  /* 50 */ {  2, 0 },  /* FICOM */
  /* 51 */ {  2, 0 },  /* FICOMP */
  /* 52 */ {  2, 0 },  /* */
  /* 53 */ {  2, 0 },  /* */
  /* 54 */ {  2, 0 },  /* */
  /* 55 */ {  2, 0 },  /* */
  /* 56 */ {  2, 0 },  /* */
  /* 57 */ {  0, 0 },  /* */
  /* 58 */ {  2, 1 },  /* */
  /* 59 */ {  2, 1 },  /* */
  /* 60 */ { 10, 0 },  /* FBLD */
  /* 61 */ {  8, 0 },  /* */
  /* 62 */ { 10, 1 },  /* FBSTP */ 
  /* 63 */ {  8, 1 }   /* */
};

unsigned _fp(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  unsigned char len, type, mod2=byte2>>6;
  unsigned char len_type;

  if (mod2==3) return no_mem(2, 0, 0);              /* reg1, reg2 */

  len_type = ((byte1&7)<<3) + ((byte2&0x3F)>>3);
  
  len  = float_len_type[len_type].len;
  type = float_len_type[len_type].type;

  if (!len) panic("float_point len==0!");

  return check_mod(0 /*imm*/, len, type);

}
