//////////////////////////////////////////////////////////////////////////////////
//
//  handlers.s 
//
//  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
//
//  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
//
//////////////////////////////////////////////////////////////////////////////////
//
// $Author: ronsse $
// $Date: 2002/02/28 12:24:13 $
// $Log: handlers.s,v $
// Revision 1.5  2002/02/28 12:24:13  ronsse
// Merged updates from jmaebe branch, except the dynlib stuff.
//
// Revision 1.4.2.7  2002/02/22 17:03:42  jmaebe
// * made checking of ret from main thread safe
// * LOOKUP_RET_HANDLER now also uses LOOKUP_CLONE2INSTR and LOOKUP_RELOC when
//   the ret address is in a dynlib
// * ret_handler (handlers.s) now saves all registers (it calls the C-routine
//   LOOKUP_RET_HANDLER), should be slightly faster and is now jumped at
//   instead of called
// * fac now works reliably to when instrumented!!
//
// Revision 1.4.2.6  2002/02/22 14:23:31  jmaebe
// + save return address of main at the start of the program and use it to
//   stop instrumenting when we return from main
//
// Revision 1.4.2.5  2002/02/20 12:20:50  jmaebe
// - removed unnecessary check for fini2 in adjust_target_
//
// Revision 1.4.2.4  2002/02/14 17:47:32  jmaebe
// + support for "lock" (merged, including fix)
// + support for cmpxchg (merged)
// * made jmp/call reg/mem again even slightly faster :)
// + support for in/out
//
// Revision 1.4.2.3  2002/02/13 15:48:25  jmaebe
// * optmized jmp_mem/call_mem and jmp_reg/call_reg handling
// * fixed interrupt_handler()
//
// Revision 1.4.2.2  2002/02/13 12:16:58  jmaebe
// + support for interrupts
// * instrumenting dynamic libraries now works, but libc still causes some
//   problems because of unsupported instructions (e.g lock and cmpxchgl)
//
// Revision 1.4.2.1  2002/02/12 16:47:20  jmaebe
// + initial implementation of instrumenting of dynamically loaded libraries,
//   but not yet working (even main(void) {} crashes :). Also, some backends
//   which have to be modified before they can work with this new feature
//   haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
//   the functionality (it's even required currently, because there's a
//   reference to ddelta (dynlib delta) in handlers.s which can't be disabled
//   using a conditional. Method used: separate clone for the dynlibs (thanks
//   for the idea, Michiel :)
//
// Revision 1.4  2001/11/14 14:04:49  ronsse
// Changed /* to // type of comments in copyright notice.
//
// Revision 1.3  2001/10/25 13:46:01  ronsse
// Initial commit
//
//
//////////////////////////////////////////////////////////////////////////////////


.include "macros.inc"
	
.text
        .align 4

/////////////////////////////////////////////////////////////////////////////////////
                
.globl get_bp
        .type get_bp,@function
get_bp: 
        movl %ebp,%eax
        ret
        .size get_bp,.-get_bp
	
/////////////////////////////////////////////////////////////////////////////////////
#ifdef INSTRUMENT_DYNLIBS

// NOTE: this routine expects to be *jumped* at, not called!!!
.globl adjust_target_
	.type adjust_target_, @function
adjust_target_:
	pushl %eax
	pushf

	// (%esp): flags, 4(%esp): eax, 8(%esp): argument (jump target)

	// target < end?

	movl  8(%esp),%eax
	cmpl  end,%eax
	jg    .Ljmpreg_dynlib
	addl  delta,%eax
	jmp   .Ljmpreg_done
.Ljmpreg_dynlib:
//	addl  ddelta,%eax
.Ljmpreg_done:
	movl  %eax,8(%esp)
	popf
	popl %eax
	ret
        .size adjust_target_,.-adjust_target_
#endif INSTRUMENT_DYNLIBS
/////////////////////////////////////////////////////////////////////////////////////

.globl jcc_taken_handler_
	.type jcc_taken_handler_, @function

jcc_taken_handler_:

	pushl %ebp
	movl %esp, %ebp
	pushal
	pushf	

	movl 4(%ebp), %eax
	// arg=IP
	pushl %eax
	call jcc_taken_handler
	leal 4(%esp), %esp
	movl %eax, 4(%ebp)
	
	popf
	popal

	leave
		
	ret
	
	.size jcc_taken_handler_,.-jcc_taken_handler_
	
/////////////////////////////////////////////////////////////////////////////////////	
		
		
.globl ret_handler_
	.type ret_handler_, @function

ret_handler_:
	
	// this handler is *jumped* at, not called -> argument is now at (%esp)	
	pushal
	movl 32(%esp), %eax
	pushf

	// NOW:	 eax=jump address

	// stop instrumenting when we return from main
	cmpl main_ret,%eax
	je  .LNoAdjust

	// check where we have to return to
	pushl %eax
	call LOOKUP_RET_ADDRESS

	// remove LOOKUP_RET_ADDRESS argument (pushed eax) from the stack
	leal 4(%esp), %esp
	
	// put modified address back on the stack (there has been a extra pushf)
	movl %eax, 36(%esp)
.LNoAdjust:
	// opkuisen & springen
	popf
	popal
	
	// ret address is now at the top of the stack
	ret
	
	.size ret_handler_,.-ret_handler_

	
/////////////////////////////////////////////////////////////////////////////////////

		
.globl trace_
        .type trace_,@function
	
trace_:
	
    //  reti 16
	
        pushl %ebp
        movl %esp, %ebp
        
        pushal
        pushf

	// 4 argumenten kopieren
	
	//  ADDRESS	
        movl 20(%ebp), %eax
        pushl %eax
	
	//  LEN
        movl 16(%ebp), %eax
        pushl %eax
	
	//  L/S
        movl 12(%ebp), %eax
        pushl %eax
	
	//  IP+delta+1
        movl 8(%ebp), %eax
        pushl %eax
	
        call *CB_data
        addl $16, %esp	
        popf
        popal
        leave
	
	// cleanup!!
        reti 16
	
        .size trace_,.-trace_

//////////////////////////////////////////////////////////////////


.globl trace_code_
        .type trace_code_,@function
	
trace_code_:   //  trace_code_(start, stop)
        
//      reti 8
	
        pushl %ebp
        movl %esp, %ebp
        
        pushal
        pushf

        // 2 argumenten kopieren
	

        movl 8(%ebp), %eax
        pushl %eax
        movl 12(%ebp), %eax
        pushl %eax
        
        call *CB_code
        addl $8, %esp
        popf
        popal
        leave
	
	// cleanup!!
        reti 8

        .size trace_code_,.-trace_code_

