/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti.h 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2002/02/28 12:24:14 $
 * $Log: jiti.h,v $
 * Revision 1.11  2002/02/28 12:24:14  ronsse
 * Merged updates from jmaebe branch, except the dynlib stuff.
 *
 * Revision 1.10  2002/02/19 13:14:12  jmaebe
 * + all 32bit registers defined (only EAX was defined until now)
 *
 * Revision 1.9  2002/02/15 14:43:54  jmaebe
 * + support for "ret imm" (needed for Pascal code :)
 *
 * Revision 1.8.2.5  2002/02/25 16:25:17  jmaebe
 * + deactivate jiti when exit() is called too (fixes crash at the end of gzip)
 *
 * Revision 1.8.2.4  2002/02/22 14:48:35  jmaebe
 * * fixed gs override support
 *
 * Revision 1.8.2.3  2002/02/22 14:23:31  jmaebe
 * + save return address of main at the start of the program and use it to
 *   stop instrumenting when we return from main
 *
 * Revision 1.8.2.2  2002/02/20 08:30:51  jmaebe
 * - removed 'ifdef/endif INSTRUMENT_DYNLIBS" defines
 * + support for "ret imm" (needed for Pascal code :) (merged r1.9)
 * + all 32bit registers defined (only EAX was defined until now) (merged r1.10)
 *
 * Revision 1.8.2.1  2002/02/12 16:47:20  jmaebe
 * + initial implementation of instrumenting of dynamically loaded libraries,
 *   but not yet working (even main(void) {} crashes :). Also, some backends
 *   which have to be modified before they can work with this new feature
 *   haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
 *   the functionality (it's even required currently, because there's a
 *   reference to ddelta (dynlib delta) in handlers.s which can't be disabled
 *   using a conditional. Method used: separate clone for the dynlibs (thanks
 *   for the idea, Michiel :)
 *
 * Revision 1.8  2002/02/05 17:20:29  jmaebe
 *   * PUSHL8 fixed (argument is a char, not a long)
 *
 * Revision 1.7  2002/01/10 21:25:30  ronsse
 * *** empty log message ***
 *
 * Revision 1.6  2001/11/12 10:11:30  ronsse
 * Added jiti_lock/jiti_unlock for backends
 *
 * Revision 1.5  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Common include file for JiTI86.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#ifndef __JITI_H__
#define __JITI_H__

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef EBUG
# define DEBUG
#endif

#ifdef DEBUG
# define DBG(X) X
#else
# define DBG(X)   
#endif

#include "dis-asm.h"
#include <pthread.h>
#include <stdlib.h>

#define TRUE  1
#define FALSE 0

/* MEMTYPE */
#define LOAD   0
#define STORE  1
#define MODIFY 2

#define TID(X)   (X+1023)/1024
#define THREADID TID(pthread_self())  

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))


// special callbacks
#define CB_EXIT (char*)1
#define CB_MEM  (char*)2
#define CB_CODE (char*)3
#define CB_INIT (char*)4

extern unsigned _instrument_code_;
extern unsigned _instrument_mem_;
extern unsigned _instrument_dyn_;

extern unsigned _optimize_;

extern unsigned tot_bp;
extern unsigned bp;
extern unsigned main_ret;
extern unsigned delta;
extern unsigned ddelta;
extern unsigned end;
extern unsigned exit_jmpaddr;
//extern unsigned addr;
//extern unsigned len;

extern FILE * jiti_tty;

extern unsigned ori_start, ori_stop, clone_start, clone_stop;

extern int calc_leal_len(char * instr, unsigned len);
extern void panic(char *);
extern char _jiti_naam[100];

//void iret_helper(void);

#define BP 0xCC   /* int 3 */

#define CALL_IMM   0xE8

// Volgende instructie flusht de reeds geprefetchte instructies.
// Dit volstaat voor self-modifying code
#define FLUSH_INSTRUCTIONS asm(".word 0x00EB")  // JMP +0

/////////////////////////////////////////////////////////

#define GET_CHAR(addr)                                  \
     (*(char*)(addr)&0xFF)

#define PUT_CHAR(addr, value)                           \
     *(char*)(addr)=(value)

#define PUT_SHORT(addr,value)                           \
     *(unsigned short*)(addr)=(value)

#define GET_SHORT(addr)                                 \
     *(unsigned short*)(addr)

#define GET_LONG(addr)                                  \
     *(unsigned*)(addr)

#define PUT_LONG(addr, value)                           \
     *(unsigned*)(addr)=(unsigned)(value)

#define PUT_COPY(addr, delta, len)                      \
     memcpy((char*)(addr), (char*)(addr)-delta, len)

/////////////////////////////////////////////////////////

#define PUT_CALL(addr, rel_target)                      \
     PUT_CHAR(addr, CALL_IMM);                          \
     PUT_LONG(addr+1, rel_target);

#define JUMP_TO(addr)                                   \
     PUT_LONG(bp+68, addr);

#define RESTART_INSTRUCTION                             \
     JUMP_TO(address); 

#define SKIP_INSTRUCTION           

/////////////////////////////////////////////////////////

#define SIZE(opcode)                                    \
     ((opcode)&1)

#define DIRECTION(opcode)                               \
     (((opcode)&2)>>1)

/////////////////////////////////////////////////////////

/* instrumenteer instructie */
#define INSTR_CALL(handler)                             \
     PUT_CALL(address, (unsigned)handler-address-5);    \
     RESTART_INSTRUCTION;

/* enkele trampoline */
#define RESTART_HANDLER(handler)                        \
     from_address=GET_LONG(bp+68);                      \
     JUMP_TO(handler);

/////

#define PUSHL32(addr, imm32)                           \
     PUT_CHAR(addr, 0x68);                             \
     PUT_LONG(addr+1, imm32);                          


#define PUSHL8(addr, imm8)                             \
     PUT_CHAR(addr, 0x6A);                             \
     PUT_CHAR(addr+1, imm8);                           


#define JMPL(addr, disp32)                             \
     PUT_CHAR(addr, 0xE9);                             \
     PUT_LONG(addr+1, disp32);

#define MOVL_EAX_2_MEM(addr, memaddr)                  \
     PUT_CHAR(addr, 0xA3);                             \
     PUT_LONG(addr+1, memaddr);                        


#define MOVL_MEM_2_EAX(addr, memaddr)                  \
     PUT_CHAR(addr, 0xA1);                             \
     PUT_LONG(addr+1, memaddr);                        


/* addl $delta,(%esp) */
#define ADDL_ESP(addr, delta)                          \
     PUT_CHAR(addr, 0x81);                             \
     PUT_CHAR(addr+1, 0x04);                           \
     PUT_CHAR(addr+2, 0x24);                           \
     PUT_LONG(addr+3, delta);                          

/* addl $delta, 4(%esp) */
#define ADDL_4ESP(addr, delta)                         \
     PUT_LONG(addr, 0x04244481);                       \
     PUT_LONG(addr+4, delta);                          

#define PUSHF                                          \
     PUT_CHAR(X_instr, 0x9C);  

#define POPF                                           \
     PUT_CHAR(X_instr, 0x9D);  

#define RET                                            \
     PUT_CHAR(X_instr, 0xC3);  

#define PUSH_REG(reg)                                  \
     PUT_CHAR(X_instr, 0x50|reg);

#define POP_REG(reg)                                   \
     PUT_CHAR(X_instr, 0x58|reg);

#define EAX 0
#define EBX 3
#define ECX 1
#define EDX 2
#define ESP 4
#define EBP 5
#define ESI 6
#define EDI 7
#define GS  8

/////////////////////////////////////////////////////////

typedef struct _clone2instr_ {
  unsigned from;
  unsigned to;
} _clone2instr_t;

extern _clone2instr_t _clone2instr_[50000];


unsigned register_callback(char * function, unsigned callback);
unsigned decode_block(unsigned address);
void process_elf_info(char * executable);
unsigned get_bp();
void par_init();
void _jiti_init_addr2line(char * _jiti_naam);
void _jiti_init_info();
unsigned intercept_plt(char * name, unsigned new_target);
unsigned LOOKUP_RELOC(unsigned address);
unsigned no_mem(unsigned, unsigned, unsigned);
void panic(char * string);
unsigned check_mod(unsigned imm_len, unsigned len, unsigned type);
void skip();
void add_to_plt(char * name, unsigned plt_address);
unsigned LOOKUP_CLONE2INSTR(unsigned);

char * jiti_program_name(void);

extern char * jiti_version;

/* Dit moet op het einde staan. Bovendien moet jiti.h als laatste
   headerfile geincludeerd worden!! */

#define INIT(backend_callback)\
 _init(){jiti_init(backend_callback);}

#define CALL_FUNCTION(XXX)\
  ((void(*)(void))(XXX))()

// voor de backends
void jiti_lock();
void jiti_unlock();

#endif //__JITI_H__
