/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  info.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2002/02/19 17:59:24 $
 * $Log: info.c,v $
 * Revision 1.8  2002/02/19 17:59:24  ronsse
 * Better color handling
 *
 * Revision 1.7  2002/01/10 21:23:18  ronsse
 * Finally decided to use print_insn_i386 against a statically linked libopcodes.a
 *
 * Revision 1.6  2001/11/13 18:09:13  ronsse
 * Now uses addr2line for retrieving debug info (instead of bfd functions).
 *
 * Revision 1.5  2001/11/12 10:07:28  ronsse
 * Now works correct if executable lacks debug info.
 *
 * Revision 1.4  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "jiti.h"
#include "dis-asm.h"


/********************* DISASSEMBLER *******************/

static disassemble_info _jiti_intern_info;

///////////////////////////////

int _jiti_intern_read_mem_func(bfd_vma memaddr, bfd_byte *myaddr, int length, struct disassemble_info *info){
  memcpy (myaddr, info->buffer + (memaddr - info->buffer_vma), length);
  return 0;
}

//////////////////////////////

void _jiti_init_info(){
  _jiti_intern_info.fprintf_func           = (fprintf_ftype)fprintf;
  _jiti_intern_info.stream                 = stdout;
  _jiti_intern_info.symbols                = NULL;
  _jiti_intern_info.num_symbols            = 0;
  _jiti_intern_info.buffer                 = 0;
  _jiti_intern_info.buffer_vma             = 0;
  _jiti_intern_info.buffer_length          = 2^31;
  _jiti_intern_info.read_memory_func       = _jiti_intern_read_mem_func;
  _jiti_intern_info.memory_error_func      = perror_memory;
  _jiti_intern_info.print_address_func     = generic_print_address;
  _jiti_intern_info.symbol_at_address_func = generic_symbol_at_address;
  _jiti_intern_info.flags                  = 0;
  _jiti_intern_info.bytes_per_line         = 0;
  _jiti_intern_info.bytes_per_chunk        = 0;
  _jiti_intern_info.display_endian         = BFD_ENDIAN_LITTLE;
  _jiti_intern_info.insn_info_valid        = 0;                                        
  _jiti_intern_info.endian                 = BFD_ENDIAN_LITTLE;
  _jiti_intern_info.mach                   = bfd_mach_i386_i386; 
  _jiti_intern_info.arch                   = bfd_arch_i386;
}

////////////////////////////

int print_insn_i386 PARAMS ((bfd_vma, disassemble_info*)) ;

int jiti_fprint_ins1(FILE * file, unsigned printaddress, unsigned address){
  int len;
  _jiti_intern_info.stream=file;
  if (printaddress) fprintf(file, "0x%.8x: ", address);
  len=print_insn_i386(address, &_jiti_intern_info);        /* libopcodes */
  return len;
}

///////////////////////////

void jiti_fprint_ins(FILE * file, unsigned start_address, unsigned stop_address){
  while (start_address<stop_address) {
    fprintf(file, "\t\t\t\t");
    start_address+=jiti_fprint_ins1(file, 1, start_address);
    fprintf(file, "\n");
  }
}


/********************* DEBUG INFO *******************/

void print_1_line(const char * file, unsigned lineno){
  unsigned i=0;
  char lijn[256];
  FILE * fp=fopen(file, "r");
  
  if (!fp) {
    printf("<file not found>");
    return;
  } else {
    lineno--;
    // we loopen tot de lijn ervoor!!!
    while (i++<lineno)
      fgets(lijn, 255, fp);
    // lijn ervoor
    printf("\t\t%d:  %s", lineno, lijn);

    // lijn zelf
    fgets(lijn, 255, fp);
    lijn[strlen(lijn)-1]=0;  // remove CR
    printf("\t\t%d:  \033[01;05;37;41m%s\033[00m\033[m\n", lineno+1, lijn);

    // lijn erna
    fgets(lijn, 255, fp);
    printf("\t\t%d:  %s", lineno+2, lijn);

    fclose(fp);
  }
  
}

////////////////////////////////////////////////////

void jiti_addr2line(unsigned address){
  char pijp[255], bestand[255], functie[255];
  int lijn;
  FILE * fin;

  sprintf(pijp, "addr2line -fe %s 0x%x | tr : \" \"", jiti_program_name(), address);
  fin=popen(pijp, "r");
  if (!fin){perror("addr2line...");exit(-1);}
  fscanf(fin, "%s", functie);
  fscanf(fin, "%s %d", bestand, &lijn);
  pclose(fin);                                                  

  if (lijn){
	  printf("%s+%u: <%s>\n", bestand, lijn, functie);
	  print_1_line(bestand, lijn);
  } else
	  printf("<no source code found>\n");
}

