
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  _get_source_lines.c
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/11/12 10:12:27 $
 * $Log: _get_source_lines.c,v $
 * Revision 1.2  2001/11/12 10:12:27  ronsse
 * Added copyright statement
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <fcntl.h>      
#include <stdio.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>     


#define MIN(X,Y) ((X)<(Y)?(X):(Y))
#define MAXLENGTE 50


typedef struct _bestandsinfo_ {
	char naam[255];
	int * lijnen;
	void * mapping;
	struct bestandsinfo * next;
} _bestandsinfo;

_bestandsinfo * bestandsinfo = NULL;


void *MALLOC(size_t size){
	void * result=malloc(size);
	if (!result) {
		perror("malloc");
		exit(-1);
	}
	return result;
}


///////////////////////////////////////////////////////


void generate(char * bestand, int lijn){
	char bronlijn[255];
	int lengte;

	_bestandsinfo * loper = bestandsinfo;

	while (loper)
		if (!strcmp(loper->naam, bestand))
			break;
		else
			loper=(_bestandsinfo*)loper->next;

	if (!loper) {
		struct stat f_stat;
		int sourcefile_aantallijnen=0;
		int sourcefile_lengte;
		int i, sourcefile;


		loper = (_bestandsinfo*)MALLOC(sizeof(_bestandsinfo));
		strcpy(loper->naam, bestand);
		loper->next=(struct bestandsinfo*)bestandsinfo;
		bestandsinfo=loper;

		fprintf(stderr, "Opening %s", bestand);

		sourcefile=open(bestand, O_RDONLY);
		if (sourcefile==-1) {
			puts("--source file not found--");
			return;
		}

		fstat(sourcefile, &f_stat);
		sourcefile_lengte=f_stat.st_size;

		fprintf(stderr, ", size=%d bytes", sourcefile_lengte);

		loper->mapping=mmap(0, sourcefile_lengte, PROT_READ, MAP_SHARED, sourcefile, 0);
		if (loper->mapping==MAP_FAILED) {perror("mapping");exit(-1);}

		close(sourcefile);

		// we tellen eerst het aantal lijnen
		for (i=0; i<sourcefile_lengte; i++)
			if (((char*)loper->mapping)[i]=='\n')
				sourcefile_aantallijnen++;

		loper->lijnen=(int*)MALLOC((sourcefile_aantallijnen+1)*sizeof(int));
		loper->lijnen[0]=0;
		sourcefile_aantallijnen=1;

		for (i=0; i<sourcefile_lengte; i++)
			if (((char*)loper->mapping)[i]=='\n')
				loper->lijnen[sourcefile_aantallijnen++]=i+1;

		fprintf(stderr, ", %d lines.\n", sourcefile_aantallijnen);

	}

	lijn--;
	lengte=MIN(loper->lijnen[lijn+1]-loper->lijnen[lijn],MAXLENGTE);
	strncpy(bronlijn, loper->mapping+loper->lijnen[lijn], lengte);
	bronlijn[lengte-1]=0;

	puts(bronlijn);
}

///////////////////////////////////////////////////////////////

int main(int argc, char ** argv){
	FILE * fin;
	char bestand[100];
	int lijn, aantal;
	char pijp[255];

	sprintf(pijp, "grep ^\\> %s | cut -b3-10 | addr2line -fse %s  | tr : \" \" | awk NF==2", argv[1], argv[2]);
	fin=popen(pijp, "r");
	if (!fin){perror("grep...");exit(-1);}

	do {
		aantal=fscanf(fin, "%s %d\n", bestand, &lijn);
		generate(bestand, lijn);
	} while (aantal==2);

	pclose(fin);

	return 0;
}
